import { MTLLoader } from 'three/examples/jsm/loaders/MTLLoader.js';
import { OBJLoader } from 'three/examples/jsm/loaders/OBJLoader.js';
import { GLTFLoader } from 'three/examples/jsm/loaders/GLTFLoader.js';
import { Group } from 'three';
import { DRACOLoader } from 'three/examples/jsm/loaders/DRACOLoader.js';

export interface IObjectLoader {
  /**
   * mtl파일이 포함된 obj 파일 형식을 로드
   * @param files
   * @param filesMap
   */
  objectLoadAsFolder(files: File[]): Promise<any>;
}

class ObjectLoaderService implements IObjectLoader {
  /**
   * mtl파일이 포함된 obj 파일 형식을 로드
   * @param files
   * @param filesMap
   */
  async objectLoadAsFolder(files: File[]) {
    const { type, textureImages, index3DFiles } = this.#getFileFormat(files);

    if (type === 'glb') {
      return await this.#glbLoad(index3DFiles);
    }

    if (type === 'mtl' || type === 'obj') {
      return await this.#mtlLoad(index3DFiles, textureImages);
    }
  }

  async #mtlLoad(index3DFiles: { [p: string]: File }, textureImages: { [p: string]: File }) {
    const mtlString = await index3DFiles['mtl'].text();
    const objString = await index3DFiles['obj'].text();
    const mtlLoader = new MTLLoader();
    const objLoader = new OBJLoader();

    const materials = mtlLoader.parse(mtlString, '');

    mtlLoader.manager.setURLModifier((url) => {
      url = url.replace(/^(\.?\/)/, ''); // remove './'

      const blob = textureImages[url];

      if (blob) {
        return URL.createObjectURL(blob);
      }

      return '/';
    });

    materials.preload();

    const object = objLoader.setMaterials(materials).parse(objString);

    return object;
  }

  async #glbLoad(index3DFiles: { [p: string]: File }): Promise<Group> {
    const gltfLoader = new GLTFLoader();
    const glbFile = index3DFiles['glb'];
    const dracoLoader = new DRACOLoader();

    dracoLoader.setDecoderPath('/draco/');
    gltfLoader.setDRACOLoader(dracoLoader);

    return new Promise((resolve, reject) => {
      glbFile.arrayBuffer().then((res) => {
        gltfLoader.parse(res, '', (gltf) => {
          resolve(gltf.scene);
        });
      });
    });
  }

  #filterFiles(files: File[]) {
    const textureImages = {} as { [p: string]: File };
    const index3DFiles = {} as { [p: string]: File };

    files.forEach((file) => {
      const splitName = file.name.split('.');
      const fileName = splitName[splitName.length - 1];

      if (fileName === 'jpg' || fileName === 'png' || fileName === 'jpeg') {
        textureImages[file.name] = file;
      }

      if (fileName === 'mtl') {
        index3DFiles['mtl'] = file;
      }

      if (fileName === 'obj') {
        index3DFiles['obj'] = file;
      }

      if (fileName === 'glb') {
        index3DFiles['glb'] = file;
      }
    });

    return { textureImages, index3DFiles };
  }

  #getFileFormat(files: File[]) {
    const { index3DFiles, textureImages } = this.#filterFiles(files);

    return { type: Object.keys(index3DFiles)[0], textureImages, index3DFiles };
  }
}

export default ObjectLoaderService;
